#pragma once

namespace AMDiS
{
  template <class MeshView>
  class BoundaryFacetIterator;

  // An Iterator over all elements and when element hasBoundaryIntersections
  template <class MeshView>
  class BoundaryElementIterator
  {
    friend class BoundaryFacetIterator<MeshView>;
    
    using Self = BoundaryElementIterator;
    using Element = typename MeshView::template Codim<0>::Entity;
    using ElementIterator = typename MeshView::template Codim<0>::Iterator;
    
    class Iterator
    {
    public:
      Iterator(ElementIterator elementIt, ElementIterator endIt)
        : elementIt(elementIt)
        , endIt(endIt)
      {}
      
      Iterator(Iterator const&) = default;
      Iterator& operator=(Iterator const&) = default;
      
      // iterate to next element that has boundary intersections
      Iterator& operator++()
      {
        ++elementIt;
        while (elementIt != endIt && !elementIt->hasBoundaryIntersections())
          ++elementIt;
        return *this;
      }
      
      Iterator operator++(int)
      {
        auto tmp = *this;
        ++(*this);
        return tmp;
      }
      
      Element const& operator*() const
      {
        return *elementIt;
      }
      
      Element const* operator->() const
      {
        return &(*elementIt);
      }
      
      bool operator==(Iterator const& that) const
      {
        return elementIt == that.elementIt;
      }
      
      bool operator!=(Iterator const& that) const
      {
        return !(*this == that);
      }
      
    private:
      ElementIterator elementIt;
      ElementIterator endIt;
    };

  public:
    /// Constructor.
    BoundaryElementIterator(MeshView const& meshView)
      : meshView(meshView)
    {}
    
    Iterator begin() {
      auto elementIt = elements(meshView).begin();
      auto endIt = elements(meshView).end();
      while (elementIt != endIt && !elementIt->hasBoundaryIntersections())
        ++elementIt;
      
      return {elementIt, endIt};
    }
    
    Iterator end() {
      return {elements(meshView).end(), elements(meshView).end()};
    }
    
  private:
    MeshView const& meshView;
  };
  
  
  /// Generator function for the boundary-element iterator
  template <class MeshView>
  BoundaryElementIterator<MeshView> boundary_elements(MeshView const& meshView)
  {
    return {meshView};
  }
  

  // An Iterator over all elements and when element hasBoundaryIntersections, then 
  // iterate over all boundary-intersections with given Index, oder for thos with 
  // predicate returns true
  template <class MeshView>
  class BoundaryFacetIterator
  {
    using Element = typename MeshView::template Codim<0>::Entity;
    using Facet   = typename MeshView::template Codim<1>::Entity;
    
    using ElementIterator = typename BoundaryElementIterator<MeshView>::Iterator;
    using FacetIterator = typename MeshView::IntersectionIterator;
    
    class Iterator
    {
    public:
      Iterator(MeshView const& meshView, 
               ElementIterator elementIt, 
               FacetIterator facetIt, 
               ElementIterator endIt)
        : meshView(meshView)
        , elementIt(elementIt)
        , facetIt(facetIt)
        , endIt(endIt)
      {}
      
      Iterator(Iterator const&) = default;
      Iterator& operator=(Iterator const&) = default;
      
      // iterate to next boundary face within current element or to the first
      // boundary face in the next boundary element
      Iterator& operator++()
      {
        ++facetIt;
        do {
          auto facetEndIt = intersections(meshView, *elementIt).end();
          while (facetIt != facetEndIt && !facetIt->boundary())
            ++facetIt;
          if (facetIt != facetEndIt)
            break;
          
          ++elementIt;
          facetIt = intersections(meshView, *elementIt).begin();
        } while (elementIt != endIt);
        
        return *this;
      }
      
      Iterator operator++(int)
      {
        auto tmp = *this;
        ++(*this);
        return tmp;
      }
      
      // returns Dune::Intersection
      auto const& operator*() const
      {
        return *facetIt;
      }
      
      auto const* operator->() const
      {
        return &(*facetIt);
      }
      
      bool operator==(Iterator const& that) const
      {
        return elementIt == that.elementIt && (elementIt == endIt || facetIt == that.facetIt);
      }
      
      bool operator!=(Iterator const& that) const
      {
        return !(*this == that);
      }
      
    private:
      MeshView const& meshView;
      ElementIterator elementIt;
      FacetIterator   facetIt;
      ElementIterator endIt;
    };

  public:
    /// Constructor.
    BoundaryFacetIterator(MeshView const& meshView)
      : meshView(meshView)
    {}
    
    Iterator begin() {
      auto elementIt = boundary_elements(meshView).begin();
      auto endElementIt = boundary_elements(meshView).end();
      auto facetIt = intersections(meshView, *elementIt).begin();
      
      while (!facetIt->boundary())
        ++facetIt;
      
      return {meshView, elementIt, facetIt, endElementIt};
    }
    
    Iterator end() {
      auto elementIt = boundary_elements(meshView).begin();
      auto endElementIt = boundary_elements(meshView).end();
      auto facetIt = intersections(meshView, *elementIt).begin();
      return {meshView, endElementIt, facetIt, endElementIt};
    }
    
  private:
    MeshView const& meshView;
  };
  
  
  /// Generator function for the boundary-element iterator
  template <class MeshView>
  BoundaryFacetIterator<MeshView> boundary_facets(MeshView const& meshView)
  {
    return {meshView};
  }

} // end namespace AMDiS
