#pragma once

#include <functional>
#include <type_traits>
#include <vector>

#include <dune/amdis/common/ConceptsBase.hpp>

#include <dune/amdis/Output.hpp>

namespace AMDiS
{
  /// Implements a boundary condition of Dirichlet-type.
  /**
   * By calling the methods \ref init() and \ref finish before and after
   * assembling the system-matrix, respectively, dirichlet boundary conditions
   * can be applied to the matrix and system vector. Therefore a predicate
   * functions indicates the DOFs where values should be enforced and a second
   * functor provided in the constructor is responsible for determining the
   * values to be set at the DOFs.
   *
   * In the \ref finish method the matrix is called with \ref applyDirichletBC
   * to erase the corresponding rows and columns for the DOF indices. This
   * application of boundary conditions can be symmetric if the matrix does
   * support this symmetric modification. As a result, this method returns a list
   * of columns values, that should be subtracted from the rhs.
   **/
  template <class WorldVector>
  class DirichletBC
  {
  public:
    template <class Predicate, class Values,
      std::enable_if_t< Concepts::Functor<Predicate, bool(WorldVector)> &&
				                Concepts::Functor<Values,  double(WorldVector)>, int*> = nullptr>
    DirichletBC(Predicate&& predicate, Values&& values)
      : predicate(std::forward<Predicate>(predicate))
      , values(std::forward<Values>(values))
    {}


    template <class Matrix, class VectorX, class VectorB>
    void init(bool apply,
              Matrix& matrix,
              VectorX& rhs,
              VectorB& solution);


    template <class Matrix, class VectorX, class VectorB>
    void finish(bool apply,
                Matrix& matrix,
                VectorX& rhs,
                VectorB& solution);

  private:
    std::function<bool(WorldVector)> predicate;
    std::function<double(WorldVector)> values;

    bool initialized = false;
    std::vector<char> dirichletNodes;
  };

} // end namespace AMDiS

#include "DirichletBC.inc.hpp"
