#pragma once

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <vector>
#include <type_traits>

#include <dune/geometry/quadraturerules.hh>

#include <dune/amdis/OperatorEvaluation.hpp>
#include <dune/amdis/common/ValueCategory.hpp>

namespace AMDiS
{

  /// Abstract base class of all operator terms
  template <class MeshView>
  class OperatorTerm
  {
  protected:
    using Codim0  = typename MeshView::template Codim<0>;
    using Element = typename Codim0::Entity;

    static constexpr int dim = MeshView::dimension;
    static constexpr int dow = MeshView::dimensionworld;

    using QuadratureRule = Dune::QuadratureRule<double, dim>;
    using PointList = std::vector<Dune::QuadraturePoint<double, dim>>;

  public:
    virtual void init(Element const& element,
                      PointList const& points) = 0;

    virtual double evalZot(size_t iq,
        Dune::FieldVector<double,1> const& test,
        Dune::FieldVector<double,1> const trial = 1.0) const = 0;

    virtual double evalFot1(size_t iq,
        Dune::FieldVector<double,1> const& test,
        Dune::FieldVector<double,dow> const& grad_trial) const = 0;

    virtual double evalFot2(size_t iq,
        Dune::FieldVector<double,dow> const& grad_test,
        Dune::FieldVector<double,1> const trial = 1.0) const = 0;

    virtual double evalSot(size_t iq,
        Dune::FieldVector<double,dow> const& grad_test,
        Dune::FieldVector<double,dow> const& grad_trial) const = 0;

    virtual int getDegree(Dune::GeometryType const& t) const = 0;
  };



  /// Base class for all operators based on expressions
  template <class MeshView, class Term, class Traits = tag::none>
  class GenericOperatorTerm
      : public OperatorTerm<MeshView>
      , public OperatorEvaluation
  {
    using Super   = OperatorTerm<MeshView>;
    using Element = typename Super::Element;
    using PointList = typename Super::PointList;

    static constexpr int dim = MeshView::dimension;
    static constexpr int dow = MeshView::dimensionworld;

  public:
    GenericOperatorTerm(Term const& term)
      : term(term)
    {}

    virtual void init(Element const& element,
                      PointList const& points) override
    {
      term.init(element, points);

      // cache term evaluation
      values.resize(points.size());
      for (size_t iq = 0; iq < points.size(); ++iq)
        values[iq] = term[iq];
    }

    virtual double evalZot(size_t iq,
        Dune::FieldVector<double,1> const& test,
        Dune::FieldVector<double,1> const trial = 1.0) const override
    {
      return this->evalZotImpl(_cat{}, _traits{}, values[iq], test, trial);
    }

    virtual double evalFot1(size_t iq,
        Dune::FieldVector<double,1> const& test,
        Dune::FieldVector<double,dow> const& grad_trial) const override
    {
      return this->evalFotImpl(_cat{}, _traits{}, values[iq], grad_trial, test);
    }

    virtual double evalFot2(size_t iq,
        Dune::FieldVector<double,dow> const& grad_test,
        Dune::FieldVector<double,1> const trial = 1.0) const override
    {
      return this->evalFotImpl(_cat{}, _traits{}, values[iq], grad_test, trial);
    }

    virtual double evalSot(size_t iq,
        Dune::FieldVector<double,dow> const& grad_test,
        Dune::FieldVector<double,dow> const& grad_trial) const override
    {
      return this->evalSotImpl(_cat{}, _traits{}, values[iq], grad_test, grad_trial);
    }

    virtual int getDegree(Dune::GeometryType const& t) const override
    {
      return term.getDegree(t);
    }

  private:
    Term term;

    using value_type = std::decay_t< decltype( std::declval<Term>()[std::declval<size_t>()] ) >;
    using _cat    = ValueCategory_t<value_type>;
    using _traits = Traits;

    std::vector<value_type> values;
  };

} // end namespace AMDiS
