#pragma once

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <tuple>
#include <string>
#include <memory>
#include <list>
#include <map>

#include <dune/istl/matrix.hh>
#include <dune/grid/common/grid.hh>

#include <dune/amdis/AdaptInfo.hpp>
#include <dune/amdis/CreatorInterface.hpp>
#include <dune/amdis/DirichletBC.hpp>
#include <dune/amdis/FileWriter.hpp>
#include <dune/amdis/Flag.hpp>
#include <dune/amdis/Initfile.hpp>
#include <dune/amdis/LinearAlgebra.hpp>
#include <dune/amdis/Mesh.hpp>
#include <dune/amdis/Operator.hpp>
#include <dune/amdis/ProblemStatBase.hpp>
#include <dune/amdis/ProblemStatTraits.hpp>
#include <dune/amdis/StandardProblemIteration.hpp>

#include <dune/amdis/common/Timer.hpp>
#include <dune/amdis/common/TupleUtility.hpp>
#include <dune/amdis/common/TypeDefs.hpp>
#include <dune/amdis/common/Utility.hpp>

namespace AMDiS
{

  template <class Traits>
  class ProblemStatSeq : public ProblemStatBase
  {
    using Self = ProblemStatSeq;

  public: // typedefs and static constants

    using FeSpaces = typename Traits::FeSpaces;
    using Mesh     = typename Traits::Mesh;
    using MeshView = typename Mesh::LeafGridView;

    using Codim0   = typename MeshView::template Codim<0>;
    using Element  = typename Codim0::Entity;

    using ElementVector = Impl::ElementVector;
    using ElementMatrix = Impl::ElementMatrix;


    /// Dimension of the mesh
    static constexpr int dim = Traits::dim;

    /// Dimension of the world
    static constexpr int dow = Traits::dimworld;

    /// Number of problem components
    static constexpr int nComponents = Traits::nComponents;


    template <size_t I>
    using FeSpace = std::tuple_element_t<I, FeSpaces>;

    using WorldVector = typename Codim0::Geometry::GlobalCoordinate;

    using SystemVectorType = SystemVector<FeSpaces>;
    using SystemMatrixType = SystemMatrix<FeSpaces>;

    using OperatorType = Operator<MeshView>;
    using LinearSolverType = LinearSolverInterface<typename SystemMatrixType::MultiMatrix,
                                                   typename SystemVectorType::MultiVector>;

  public:
    /**
     * \brief Constructor. Takes the name of the problem that is used to
     * access values correpsonding to this püroblem in the parameter file.
     *
     * Parameters read by ProblemStatSeq, with name 'PROB'
     *   PROB->names: \ref componentNames
     **/
    ProblemStatSeq(std::string name)
      : name(name)
    {
      Parameters::get(name + "->names", componentNames);
      for (size_t i = componentNames.size(); i < nComponents; ++i)
        componentNames.push_back("solution[" + std::to_string(i) + "]");
    }


    /**
     * \brief Initialisation of the problem.
     *
     * Parameters read in initialize() for problem with name 'PROB'
     *   MESH[0]->global refinements:  nr of initial global refinements
     **/
    void initialize(Flag initFlag,
		    Self* adoptProblem = NULL,
		    Flag adoptFlag = INIT_NOTHING);


    /// Adds an operator to \ref A.
    /** @{ */
    void addMatrixOperator(OperatorType& op,
			   int i, int j,
                           double* factor = NULL,
                           double* estFactor = NULL);

    void addMatrixOperator(shared_ptr<OperatorType> op,
                           int i, int j,
                           double* factor = NULL,
                           double* estFactor = NULL);

    void addMatrixOperator(std::map< std::pair<int,int>,
                                     shared_ptr<OperatorType> > ops);
    void addMatrixOperator(std::map< std::pair<int,int>,
                                     std::pair<shared_ptr<OperatorType>, bool> > ops);
    /** @} */


    /// Adds an operator to \ref rhs.
    /** @{ */
    void addVectorOperator(OperatorType& op,
                           int i,
                           double* factor = NULL,
                           double* estFactor = NULL);

    void addVectorOperator(shared_ptr<OperatorType> op,
                           int i,
                           double* factor = NULL,
                           double* estFactor = NULL);

    void addVectorOperator(std::map< int,
                                     shared_ptr<OperatorType> > ops);
    void addVectorOperator(std::map< int,
                                     std::pair<shared_ptr<OperatorType>, bool> > ops);
    /** @} */

    /// Adds a Dirichlet boundary condition
    template <class Predicate, class Values>
    void addDirichletBC(Predicate const& predicate,
			int row, int col,
			Values const& values);


    /// Implementation of \ref ProblemStatBase::solve
    virtual void solve(AdaptInfo& adaptInfo,
		       bool createMatrixData = true,
		       bool storeMatrixData = false) override;


    /// Implementation of \ref ProblemStatBase::buildAfterCoarse
    virtual void buildAfterCoarsen(AdaptInfo& adaptInfo,
				   Flag flag,
				   bool asmMatrix = true,
				   bool asmVector = true) override;


    /// Writes output files.
    void writeFiles(AdaptInfo& adaptInfo, bool force = false);

  public: // get-methods

    /// Returns nr of components \ref nComponents
    static constexpr int getNumComponents()
    {
      return nComponents;
    }


    /// Return the \ref systemMatrix
    auto getSystemMatrix()       { return systemMatrix; }
    auto getSystemMatrix() const { return systemMatrix; }


    /// Return the \ref solution
    auto getSolution()       { return solution; }
    auto getSolution() const { return solution; }

    /// Return the i'th \ref solution component
    template <size_t I = 0>
    decltype(auto) getSolution(const index_<I> _i = {})
    {
      return (*solution)[_i];
    }


    /// Return the \ref rhs
    auto getRhs()       { return rhs; }
    auto getRhs() const { return rhs; }


    /// Return the \ref linearSolver
    auto getSolver() { return linearSolver; }

    void setSolver(std::shared_ptr<LinearSolverType> const& solver_)
    {
      linearSolver = solver_;
    }

    /// Return the \ref mesh
    auto getMesh() { return mesh; }

    void setMesh(std::shared_ptr<Mesh> const& mesh_)
    {
      mesh = mesh_;
      meshView = make_shared<MeshView>(mesh->leafGridView());

      createFeSpaces();
      createMatricesAndVectors();
      createFileWriter();
    }


    /// Return the \ref meshView
    auto getMeshView() { return meshView; }

    /// Return the \ref feSpaces
    auto getFeSpaces() { return feSpaces; }


    /// Return the I'th \ref feSpaces component
    template <size_t I = 0>
    FeSpace<I> const& getFeSpace(const index_<I> = {}) const
    {
      return std::get<I>(*feSpaces);
    }

    /// Implementation of \ref ProblemStatBase::getName
    virtual std::string getName() const override
    {
      return name;
    }

    /// Return a vector of names of the problem components
    std::vector<std::string> getComponentNames() const
    {
      return componentNames;
    }

  protected: // initialization methods

    void createMesh()
    {
      meshName = "";
      Parameters::get(name + "->mesh", meshName);
      test_exit(!meshName.empty(),
                "No mesh name specified for '", name, "->mesh'!");

      mesh = MeshCreator<Mesh>::create(meshName);
      meshView = make_shared<MeshView>(mesh->leafGridView());

      msg("Create mesh:");
      msg("#elements = "   , mesh->size(0));
      msg("#faces/edges = ", mesh->size(1));
      msg("#vertices = "   , mesh->size(dim));
      msg("");
    }

    void createFeSpaces()
    {
      feSpaces = make_shared<FeSpaces>(constructTuple<FeSpaces>(*meshView));
    }

    void createMatricesAndVectors()
    {
      systemMatrix = make_shared<SystemMatrixType>(*feSpaces);
      solution = make_shared<SystemVectorType>(*feSpaces, componentNames);

      auto rhsNames = std::vector<std::string>(nComponents, "rhs");
      rhs = make_shared<SystemVectorType>(*feSpaces, rhsNames);
    }

    void createSolver()
    {
      std::string solverName = "cg";
      Parameters::get(name + "->solver->name", solverName);

      auto solverCreator
        = named(CreatorMap<LinearSolverType>::getCreator(solverName, name + "->solver->name"));

      linearSolver = solverCreator->create(name + "->solver");
    }

    void createFileWriter()
    {
      filewriter = make_shared<FileWriter<Traits>>(name + "->output",
                                                        *meshView,
                                                        componentNames);
    }

  protected: // sub-methods to assemble DOFMatrix

    template <class LhsData, class RhsData, class Elements>
    void assemble(LhsData lhs, RhsData rhs, Elements const& elements);

    template <class RowView, class ColView>
    bool getElementMatrix(RowView const& rowView,
			  ColView const& colView,
			  ElementMatrix& elementMatrix,
			  std::list<shared_ptr<OperatorType>>& operators,
                          std::list<double*> const& factors);

    template <class RowView>
    bool getElementVector(RowView const& rowView,
			  ElementVector& elementVector,
			  std::list<shared_ptr<OperatorType>>& operators,
                          std::list<double*> const& factors);


    template <class Matrix, class RowIndexSet, class ColIndexSet>
    void addElementMatrix(Matrix& matrix,
			  RowIndexSet const& rowIndexSet,
			  ColIndexSet const& colIndexSet,
			  ElementMatrix const& elementMatrix);

    template <class Vector, class IndexSet>
    void addElementVector(Vector& vector,
			  IndexSet const& indexSet,
			  ElementVector const& elementvector);

  private: // some internal methods

    template <size_t I = 0>
    typename FeSpace<I>::NodeFactory&
    getNodeFactory(const index_<I> = index_<I>())
    {
      return const_cast<typename FeSpace<I>::NodeFactory&>(std::get<I>(*feSpaces).nodeFactory());
    }


  private:
    /// Name of this problem.
    std::string name;

    /// Stores the names for all components. Is used for naming the solution
    /// vectors, \ref solution.
    std::vector<std::string> componentNames;

    /// Mesh of this problem.
    shared_ptr<Mesh> mesh; // TODO: generalize to multi-mesh problems

    /// Name of the mesh
    std::string meshName;

    /// A gridView object
    shared_ptr<MeshView> meshView;

    /// Pointer to the meshes for the different problem components
    std::vector<Mesh*> componentMeshes;

    /// FE spaces of this problem.
    shared_ptr<FeSpaces> feSpaces; // eventuell const

    /// A FileWriter object
    shared_ptr<FileWriter<Traits>> filewriter;

    /// An object of the linearSolver Interface
    shared_ptr<LinearSolverType> linearSolver;

    /// A block-matrix that is filled during assembling
    shared_ptr<SystemMatrixType> systemMatrix;

    /// A block-vector with the solution components
    shared_ptr<SystemVectorType> solution;

    /// A block-vector (load-vector) corresponding to the right.hand side
    /// of the equation, filled during assembling
    shared_ptr<SystemVectorType> rhs;


    template <class T>
    using MatrixEntries = std::map< std::pair<int,int>, std::list<T> >;

    template <class T>
    using VectorEntries = std::map< int, std::list<T> >;


    /// A map (i,j) -> list<DirichleBC> string a boundary conditions for
    /// each matrix block
    MatrixEntries<shared_ptr<DirichletBC<WorldVector>>> dirichletBc;

    /// A map (i,j) -> list<OperatorType> string the differential operators for
    /// each matrix block
    MatrixEntries<shared_ptr<OperatorType>> matrixOperators;
    MatrixEntries<double*>                  matrixFactors;
    std::map< std::pair<int,int>, bool >    matrixAssembled; // if false, do reassemble
    std::map< std::pair<int,int>, bool >    matrixChanging;  // if true, or vectorAssembled false, do reassemble

    /// A map (i) -> list<OperatorType> string the differential operators for
    /// each rhs block
    VectorEntries<shared_ptr<OperatorType>> vectorOperators;
    VectorEntries<double*>                  vectorFactors;
    std::map< int, bool >                   vectorAssembled; // if false, do reassemble
    std::map< int, bool >                   vectorChanging;  // if true, or vectorAssembled false, do reassemble

    template <int r, int c>
    struct MatrixData
    {
      using DOFMatrixType =
        std::tuple_element_t<c, std::tuple_element_t<r, typename SystemMatrixType::DOFMatrices>>;

      DOFMatrixType&                       matrix;
      std::list<shared_ptr<OperatorType>>& operators;
      std::list<double*> const&            factors;
      bool                                 assemble;

      std::pair<int,int> row_col = {r, c};
    };

    template <int r>
    struct VectorData
    {
      using DOFVectorType =
        std::tuple_element_t<r, typename SystemVectorType::DOFVectors>;

      DOFVectorType&                       vector;
      std::list<shared_ptr<OperatorType>>& operators;
      std::list<double*> const&            factors;
      bool                                 assemble;

      int row = r;
    };
  };


#ifndef AMDIS_NO_EXTERN_PROBLEMSTAT
//   extern template class ProblemStatSeq<ProblemStatTraits<1>>;
  extern template class ProblemStatSeq<ProblemStatTraits<2,2,1>>;
  extern template class ProblemStatSeq<ProblemStatTraits<2,2,2>>;
//   extern template class ProblemStatSeq<ProblemStatTraits<3>>;
#endif

  namespace Impl
  {
    template <class ProblemStatType>
    struct ProblemStat : public ProblemStatType,
                         public StandardProblemIteration
    {
      using ProblemStatType::getName;

      /// Constructor
      explicit ProblemStat(std::string nameStr)
        : ProblemStatType(nameStr)
        , StandardProblemIteration(dynamic_cast<ProblemStatBase&>(*this))
      {}

      /**
       * \brief Determines the execution order of the single adaption steps.
       *
       * If adapt is true, mesh adaption will be performed. This allows to avoid
       * mesh adaption, e.g. in timestep adaption loops of timestep adaptive
       * strategies.
       *
       * Implementation of \ref StandardProblemIteration::oneIteration.
       **/
      virtual Flag oneIteration(AdaptInfo& adaptInfo, Flag toDo = FULL_ITERATION) override
      {
        for (size_t i = 0; i < ProblemStatType::getNumComponents(); ++i)
          if (adaptInfo.spaceToleranceReached(i))
            adaptInfo.allowRefinement(false, i);
          else
            adaptInfo.allowRefinement(true, i);

        return StandardProblemIteration::oneIteration(adaptInfo, toDo);
      }

      /// Implementation of \ref ProblemStatBase::buildBeforeRefine.
      virtual void buildBeforeRefine(AdaptInfo&, Flag) override { /* Does nothing. */ }

      /// Implementation of \ref ProblemStatBase::buildBeforeCoarsen.
      virtual void buildBeforeCoarsen(AdaptInfo&, Flag) override { /* Does nothing. */ }

      /// Implementation of \ref ProblemStatBase::estimate.
      virtual void estimate(AdaptInfo& adaptInfo) override { /* Does nothing. */ }

      /// Implementation of \ref ProblemStatBase::refineMesh.
      virtual Flag refineMesh(AdaptInfo& adaptInfo) override { return {0}; }

      /// Implementation of \ref ProblemStatBase::coarsenMesh.
      virtual Flag coarsenMesh(AdaptInfo& adaptInfo) override { return {0}; }

      /// Implementation of \ref ProblemStatBase::markElements.
      virtual Flag markElements(AdaptInfo& adaptInfo) override { return {0}; }
    };

  } // end namespace Impl

  template <class Traits>
  using ProblemStat = Impl::ProblemStat<ProblemStatSeq<Traits>>;

} // end namespace AMDiS

#include "ProblemStat.inc.hpp"

