/** \file traits.hpp */

#pragma once

// std c++ headers
#include <type_traits>

// AMDiS headers
#include <dune/amdis/common/ConceptsBase.hpp>
#include <dune/amdis/common/Mpl.hpp>
#include <dune/amdis/common/Utility.hpp>

namespace AMDiS
{  
  // some traits to test for binary operators on types
  namespace traits
  {
    namespace Impl
    {
      struct PlusOp
      {
        template <class T, class U>
        auto operator()(T t, U u) const { return t + u; }
      };

      struct MultipliesOp
      {
        template <class T, class U>
        auto operator()(T t, U u) const { return t * u; }
      };

    } // end namespace Impl

    template <class T, class U = T>
    using IsAddable = bool_<Concepts::Callable<Impl::PlusOp, T, U>>;

    template <class T, class U = T>
    using IsMultiplicable = bool_<Concepts::Callable<Impl::MultipliesOp, T, U>>;

    
    template <class T>
    using IsTriviallyCopyable = std::is_pod<T>;

    // -------------------------------------------------------------------------
    
    // larger types
    template <class... Ts>
    struct LargerType;

    template <class T1, class T2, class... Ts>
    struct LargerType<T1, T2, Ts...>
    {
      using type = typename std::conditional_t< (sizeof(T1) > sizeof(T2)),
            LargerType<T1,Ts...>,
            LargerType<T2,Ts...> >::type;
    };

    template <class T1, class T2>
    struct LargerType<T1, T2>
    {
      using type = std::conditional_t<(sizeof(T1) > sizeof(T2)), T1, T2>;
    };

    template <class T> struct LargerType<T>
    {
      using type = T;
    };

    // maximal size type
    template <class... Es>
    using MaxSizeType = typename LargerType<Size_t<Es>...>::type;

    // -------------------------------------------------------------------------
    
    namespace Impl
    {
      template <class A, class B>
      struct is_compatible 
        : std::is_same<Decay_t<A>, Decay_t<B>> {};

      template <class A, class B>
      struct is_compatible<Types<A>, Types<B>>
        : is_compatible<A,B> {};

      template <>
      struct is_compatible<Types<>, Types<>> : std::true_type {};

      template <class A0, class... As, class B0, class... Bs>
      struct is_compatible<Types<A0,As...>, Types<B0,Bs...>>
        : and_<is_compatible<A0, B0>, is_compatible<Types<As...>, Types<Bs...>>> {};
        
    } // end namespace Impl

    template <class A, class B>
    using IsCompatible = Impl::is_compatible<A,B>;

  } // end namespace traits
  
  namespace Concepts
  {
    /** \addtogroup Concepts
     *  @{
     **/
    
    /// Types support \f$ A + B \f$
    template <class A, class B>
    constexpr bool Addable = traits::IsAddable<A, B>::value;
    
    /// Types support \f$ A \cdot B \f$
    template <class A, class B>
    constexpr bool Multiplicable = traits::IsMultiplicable<A, B>::value;
    
    /// Types are the same, up to decay of qualifiers
    template <class A, class B>
    constexpr bool Compatible = traits::IsCompatible<A, B>::value;
    
    /** @} **/
    
  } // end namespace Concepts
  
} // end namespace AMDiS
