#pragma once

#include <dune/istl/preconditioners.hh>

namespace AMDiS
{

  // creators for preconditioners, depending on matrix-type
  // ---------------------------------------------------------------------------

  template <class Matrix, class VectorX, class VectorB = VectorX>
  class ISTLPreconCreator;

  template <class... MatrixRows, class VectorX, class VectorB>
  class ISTLPreconCreator<Dune::MultiTypeBlockMatrix<MatrixRows...>, VectorX, VectorB>
  {
    using BasePreconditioner = Dune::Preconditioner<VectorX, VectorB>;

  public:
    ISTLPreconCreator(std::string prefix)
      : prefix(prefix)
    {}

    template <class Matrix>
    unique_ptr<BasePreconditioner> create(std::string /*name*/, Matrix const& A)
    {
      double w = 1.0;
      Parameters::get(prefix + "->relaxation", w);

      warning("Only Richardson preconditioner available for MultiTypeBlockMatrix!");
      return make_unique<Dune::Richardson<VectorX, VectorB>>(w);
    }

  private:
    std::string prefix;
  };

  template <class Block, class Alloc, class VectorX, class VectorB>
  class ISTLPreconCreator<Dune::BCRSMatrix<Block,Alloc>, VectorX, VectorB>
  {
    using BasePreconditioner = Dune::Preconditioner<VectorX, VectorB>;

  public:
    ISTLPreconCreator(std::string prefix)
      : prefix(prefix)
    {}

    template <class Matrix>
    unique_ptr<BasePreconditioner> create(std::string name, Matrix const& A)
    {
      double w = 1.0;
      Parameters::get(prefix + "->relaxation", w);

      if (name == "diag" || name == "jacobi")
        return make_unique<Dune::SeqJac<Matrix, VectorX, VectorB>>(A, 1, w);
      else if (name == "gs" || name == "gauss_seidel")
        return make_unique<Dune::SeqGS<Matrix, VectorX, VectorB>>(A, 1, w);
      else if (name == "sor")
        return make_unique<Dune::SeqSOR<Matrix, VectorX, VectorB>>(A, 1, w);
      else if (name == "ssor")
        return make_unique<Dune::SeqSSOR<Matrix, VectorX, VectorB>>(A, 1, w);
      else if (name == "ilu" || name == "ilu0")
        return make_unique<Dune::SeqILU0<Matrix, VectorX, VectorB>>(A, w);
      else
        return make_unique<Dune::Richardson<VectorX, VectorB>>(w);
    }

  private:
    std::string prefix;
  };

} // end namespace AMDiS
