#pragma once

#include <vector>

#include <boost/numeric/mtl/matrix/mapped_inserter.hpp>

#include <dune/amdis/linear_algebra/mtl/BlockMTLMatrix.hpp>
#include <dune/amdis/linear_algebra/mtl/SystemMatrix.hpp>

namespace AMDiS
{

  /**
   * \brief BaseClass for all mapper types
   *
   * A mapper assigned a local row/column index of a cell of
   * a block matrix to a global matrix index.
   * Call \ref setRow and \ref setCol first, to select a block
   * in the block matrix. Then get with \ref row, respective \ref col,
   * the global matrix index to the assigned local index.
   **/
  template <class Derived>
  struct MapperBase
  {
    using size_type = std::size_t;

    template <class BaseInserter>
    struct Inserter {
      using type = mtl::mat::mapped_inserter<BaseInserter, Derived>;
    };

    /// set the current block row
    void setRow(size_type rb) { self().setRow(rb); }

    /// set the current block columns
    void setCol(size_type cb) { self().setCol(cb); }

    /// return global matrix row, for local row in the current matrix block
    size_type row(size_type r) const { return self().row(r); }

    /// return global matrix column, for local column in the current matrix block
    size_type col(size_type c) const { return self().col(c); }

    /// return overall number of rows
    size_type getNumRows() const { return self().getNumRows(); }

    /// return overall number of columns
    size_type getNumCols() const { return self().getNumCols(); }

    size_type getNumRows(size_type comp) const { return self().getNumRows(comp); }
    size_type getNumCols(size_type comp) const { return self().getNumCols(comp); }

    /// return number of components/blocks
    size_type getNumComponents() const { return self().getNumComponents(); }

    size_type row(size_type rb, size_type cb, size_type r)
    {
      setRow(rb);
      setCol(cb);
      return row(r);
    }

    size_type col(size_type rb, size_type cb, size_type c)
    {
      setRow(rb);
      setCol(cb);
      return col(c);
    }

    Derived&       self()       { return static_cast<Derived&>(*this); }
    Derived const& self() const { return static_cast<Derived const&>(*this); }
  };


  /// Mapper implementation for non-parallel block matrices
  struct BlockMapper : public MapperBase<BlockMapper>
  {
    /// Default constructor
    BlockMapper() = default;
    BlockMapper(BlockMapper const& other) = default;
    BlockMapper& operator=(BlockMapper const& other) = default;

    /// Constructor for BlockMTLMatrix
    template <class Value, size_t _N>
    BlockMapper(BlockMTLMatrix<Value, _N, _N> const& mat)
      : nComp(_N)
      , rowOffset(0), colOffset(0)
      , nrow(num_rows(mat)), ncol(nrow)
      , sizes(nComp)
    {
      for (size_t i = 0; i < _N; ++i)
        sizes[i] = num_rows(mat[i][i]);
    }

    /// Constructor for SystemMatrix
    template <class FeSpaces, class Value>
    BlockMapper(SystemMatrix<FeSpaces, Value> const& mat)
      : BlockMapper(mat.getMatrix())
    {}

    /// Constructor for single DOFMatrix
    template <class FeSpace, class Value>
    BlockMapper(DOFMatrix<FeSpace, Value> const& mat)
      : nComp(1)
      , rowOffset(0), colOffset(0)
      , nrow(0), ncol(0)
      , sizes(nComp)
    {
      sizes[0] = mat.N();
      nrow += sizes[0];
      ncol = nrow;
    }

    /// Constructor for system with equal components
    BlockMapper(size_type nComp, size_type nDOFperComp)
    : nComp(nComp),
      rowOffset(0), colOffset(0),
      nrow(nComp*nDOFperComp), ncol(nrow),
      sizes(nComp)
    {
       for (size_type i = 0; i < nComp; ++i)
         sizes[i] = nDOFperComp;
    }

    /// calculate row offset for row component \param r
    void setRow(size_type rb)
    {
      test_exit_dbg(rb <= sizes.size(), "row nr out of range!");
      rowOffset = sum(rb);
    }

    /// calculate column offset for col component \param c
    void setCol(size_type cb)
    {
      test_exit_dbg(cb <= sizes.size(), "column nr out of range!");
      colOffset = sum(cb);
    }

    size_type row(size_type r) const { return r + rowOffset; }
    size_type col(size_type c) const { return c + colOffset; }

    size_type getNumRows() const { return nrow; }
    size_type getNumCols() const { return ncol; }

    size_type getNumRows(size_type comp) const { return sizes[comp]; }
    size_type getNumCols(size_type comp) const { return sizes[comp]; }

    size_type getNumComponents() const { return nComp; }

  private: // methods

    ///compute the sum of sizes from [0, end)
    size_type sum(size_type end) const
    {
      size_type ret = 0;
      for (size_type i = 0; i < end; ++i)
	ret += sizes[i];
      return ret;
    }

  private: // variables

    size_type nComp = 0;
    size_type rowOffset = 0;
    size_type colOffset = 0;
    size_type nrow = 0;
    size_type ncol = 0;

    std::vector<size_type> sizes;
  };


  /// Mapper implementation for non-parallel rectangular block matrices
  struct RectangularMapper : public MapperBase<RectangularMapper>
  {
    /// Constructor for block-matrices
    template <class Value, size_t _N, size_t _M>
    RectangularMapper(BlockMTLMatrix<Value, _N, _M> const& mat)
      : nRowComp(_N), nColComp(_M)
      , rowOffset(0), colOffset(0)
      , nrow(num_rows(mat)), ncol(num_cols(mat))
      , sizes_rows(nRowComp), sizes_cols(nColComp)
    {
      for (size_type i = 0; i < _N; ++i)
        sizes_rows[i] = num_rows(mat[i][0]);
      for (size_type j = 0; j < _M; ++j)
        sizes_cols[j] = num_cols(mat[0][j]);
    }

    /// calculate row offset for row component \param r
    void setRow(size_type rb)
    {
      test_exit_dbg(rb <= sizes_rows.size(), "row nr out of range!");
      rowOffset = sum(rb, sizes_rows);
    }

    /// calculate column offset for col component \param c
    void setCol(size_type cb)
    {
      test_exit_dbg(cb <= sizes_cols.size(), "column nr out of range!");
      colOffset = sum(cb, sizes_cols);
    }

    size_type row(size_type r) const { return r + rowOffset; }
    size_type col(size_type c) const { return c + colOffset; }

    size_type getNumRows() const { return nrow; }
    size_type getNumCols() const { return ncol; }

    size_type getNumRows(size_type comp) const { return sizes_rows[comp]; }
    size_type getNumCols(size_type comp) const { return sizes_cols[comp]; }

    size_type getNumComponents() const { return nRowComp; }
    size_type getNumRowComponents() const { return nRowComp; }
    size_type getNumColComponents() const { return nColComp; }

  private: // methods

    ///compute the sum of sizes from [0, end)
    size_type sum(size_type end, std::vector<size_type> const& sizes) const
    {
      size_type ret = 0;
      for (size_type i = 0; i < end; ++i)
	ret += sizes[i];
      return ret;
    }

  private: // variables

    size_type nRowComp = 0;
    size_type nColComp = 0;
    size_type rowOffset = 0;
    size_type colOffset = 0;
    size_type nrow = 0;
    size_type ncol = 0;

    std::vector<size_type> sizes_rows;
    std::vector<size_type> sizes_cols;
  };

} // end namespace AMDiS
