#pragma once

#include <vector>
#include <string>
#include <memory>
#include <tuple>

#include <boost/numeric/mtl/matrix/dense2D.hpp>
#include <boost/numeric/mtl/matrix/compressed2D.hpp>

#include <dune/amdis/Output.hpp>
#include <dune/amdis/common/TupleUtility.hpp>
#include <dune/amdis/common/Utility.hpp>
#include <dune/amdis/linear_algebra/mtl/BlockMTLMatrix.hpp>
#include <dune/amdis/linear_algebra/mtl/DOFMatrix.hpp>

// for explicit instantiation
#include <dune/amdis/ProblemStatTraits.hpp>

namespace AMDiS
{
  /// \cond HIDDEN_SYMBOLS
  namespace Impl
  {
    // DOFMatrices = std::tuple< std::tuple< DOFMatrix<RowFeSpace, ColFeSpace, ValueType>... >... >
    template <class DOFMatrices>
    struct BuildDOFMatrices
    {
      template <size_t R>
      using DOFMatrixRow = std::tuple_element_t<R, DOFMatrices>;

      template <size_t R, size_t C>
      using DOFMatrixRowCol = std::tuple_element_t<C, DOFMatrixRow<R>>;

      // add arg to repeated constructor argument list
      template <size_t... R, class RowFeSpaces, class ColFeSpaces, class MultiMatrix>
      static DOFMatrices make(Indices<R...>,
                              RowFeSpaces&& rowFeSpaces,
                              ColFeSpaces&& colFeSpace,
                              MultiMatrix&& multiMatrix)
      {
        return DOFMatrices{make_row(
          index_<R>(),
          MakeSeq_t<std::tuple_size<std::decay_t<ColFeSpaces>>::value>(),
          std::get<R>(std::forward<RowFeSpaces>(rowFeSpaces)),
          std::forward<ColFeSpaces>(colFeSpace),
          std::forward<MultiMatrix>(multiMatrix)
          )...};
      }

      template <size_t R, size_t... C, class RowFeSpace, class ColFeSpaces, class MultiMatrix>
      static DOFMatrixRow<R> make_row(index_<R>,
                                      Indices<C...>,
                                      RowFeSpace&& rowFeSpace,
                                      ColFeSpaces&& colFeSpace,
                                      MultiMatrix&& multiMatrix)
      {
        return DOFMatrixRow<R>{DOFMatrixRowCol<R,C>(
          std::forward<RowFeSpace>(rowFeSpace),
          std::get<C>(std::forward<ColFeSpaces>(colFeSpace)),
          "matrix[" + std::to_string(R) + "][" + std::to_string(C) + "]",
          multiMatrix(index_<R>(), index_<C>())
          )...};
      }
    };

    // construction method to construct a tuple of tuples of DOFMatrices
    template <class DOFMatrices, class RowFeSpaces, class ColFeSpaces, class MultiMatrix>
    DOFMatrices buildDOFMatrices(RowFeSpaces&& rowFeSpaces,
                                  ColFeSpaces&& colFeSpace,
                                  MultiMatrix&& multiMatrix)
    {
        return BuildDOFMatrices<DOFMatrices>::make(
            MakeSeq_t<std::tuple_size<std::decay_t<RowFeSpaces>>::value>(), // traverse rows first
            std::forward<RowFeSpaces>(rowFeSpaces),
            std::forward<ColFeSpaces>(colFeSpace),
            std::forward<MultiMatrix>(multiMatrix));
    }

  } // end namespace Impl
  /// \endcond


  /// \brief Container that repesents multiple data-matrices.
  /**
    *  Represents a fixed-size matrix of \ref mtl::compressed2D + a tuple of corresponding
    *  feSpaces. This (R,C)'th matrix combined with the (R,C)'th FeSpace
    *  builds a \ref DOFMatrix and can be return by the \ref operator().
    **/
  template <class FeSpaces, class ValueType = double>
  class SystemMatrix
  {
    template <class RowFeSpace>
    struct MultiMatrixRowGenerator
    {
      template <class ColFeSpace>
      using DOFMatrixGenerator = DOFMatrix<RowFeSpace, ColFeSpace, ValueType>;
      using DOFMatrices  = MakeTuple_t<DOFMatrixGenerator, FeSpaces>;
    };

    template <class RowFeSpace>
    using DOFMatrixGenerator = typename MultiMatrixRowGenerator<RowFeSpace>::DOFMatrices;

  public:
    /// The number of blocks per row/column
    static constexpr size_t nComponent = std::tuple_size<FeSpaces>::value;
    AMDIS_STATIC_ASSERT( nComponent > 0 );

    /// The type of the matrix in each block
    using BaseMatrix = mtl::compressed2D<ValueType>;

    /// The type of the matrix of DOFMatrices
    using DOFMatrices = MakeTuple_t<DOFMatrixGenerator, FeSpaces>;

    /// The type of the matrix of base-matrices
    using MultiMatrix = BlockMTLMatrix<BaseMatrix, nComponent, nComponent>;


    /// Constructor that constructs new base-matrices
    SystemMatrix(FeSpaces const& feSpaces)
      : feSpaces(feSpaces)
      , matrix{}
      , dofmatrices(Impl::buildDOFMatrices<DOFMatrices>(feSpaces, feSpaces, matrix))
    {}


    /// Return the number of blocks per row
    static constexpr size_t N()
    {
      return std::tuple_size<FeSpaces>::value;
    }

    /// Return the number of blocks per column
    static constexpr size_t M()
    {
      return std::tuple_size<FeSpaces>::value;
    }

    /// Return the (R,C)'th underlaying base matrix
    template <size_t R, size_t C>
    auto& getMatrix(const index_<R> _r = {},
                    const index_<C> _c = {})
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return matrix(_r, _c);
    }

    /// Return the (R,C)'th underlaying base matrix
    template <size_t R, size_t C>
    auto const& getMatrix(const index_<R> _r = {},
                          const index_<C> _c = {}) const
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return matrix(_r, _c);
    }

    /// Return the underlying multi matrix
    MultiMatrix&        getMatrix()       { return matrix; }
    MultiMatrix const&  getMatrix() const { return matrix; }

    /// Return the I'th row finite element space
    template <size_t I = 0>
    auto& getRowFeSpace(const index_<I> = {}) const
    {
      static_assert(I < N(), "Index out of range [0,N)");
      return std::get<I>(feSpaces);
    }

    /// Return the I'th column finite element space
    template <size_t I = 0>
    auto const& getColFeSpace(const index_<I> = {}) const
    {
      static_assert(I < M(), "Index out of range [0,M)");
      return std::get<I>(feSpaces);
    }

    /// Return the (R,C)'th DOFMatrix
    template <size_t R = 0, size_t C = 0>
    auto& operator()(const index_<R> = {},
                     const index_<C> = {})
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return std::get<C>(std::get<R>(dofmatrices));
    }

    /// Return the (R,C)'th DOFMatrix
    template <size_t R = 0, size_t C = 0>
    auto const& operator()(const index_<R> = {},
                           const index_<C> = {}) const
    {
      static_assert(R < N() && C < M(), "Indices out of range [0,N)x[0,M)");
      return std::get<C>(std::get<R>(dofmatrices));
    }

  private:
    /// a tuple of feSpaces for rows and columns of the matrix
    FeSpaces const& feSpaces;

    /// a tuple of data-matrices
    MultiMatrix matrix;

    /// a tuple of DOFMatrices
    DOFMatrices dofmatrices;
  };

#ifndef AMDIS_NO_EXTERN_SYSTEMMATRIX
    // explicit instantiation in SystemMatrix.cpp
    extern template class SystemMatrix<typename ProblemStatTraits<2,2,1,2>::FeSpaces>;
#endif

} // end namespace AMDiS
