#pragma once

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_UMFPACK

#include <iostream>

#include <boost/numeric/mtl/operation/two_norm.hpp>
#include <boost/numeric/mtl/interface/umfpack_solve.hpp>

#include <dune/amdis/linear_algebra/RunnerInterface.hpp>
#include <dune/amdis/linear_algebra/SolverInfo.hpp>

#include <dune/amdis/linear_algebra/mtl/Copy.hpp>

namespace AMDiS
{
  /**
   * \ingroup Solver
   * \class AMDiS::UmfpackRunner
   * \brief \implements RunnerInterface
   * Wrapper for the external
   *   [UMFPACK solver](http://faculty.cse.tamu.edu/davis/suitesparse.html)
   *
   * This is a direct solver for large sparse matrices.
   */
  template <class Matrix, class Vector>
  class UmfpackRunner;

  template <class Matrix, class Vector>
  class UmfpackRunnerBase : public RunnerInterface<Matrix, Vector, Vector>
  {
  protected:
    using Super = RunnerInterface<Matrix, Vector>;
    using PreconBase = typename Super::PreconBase;

    using FullMatrix = BaseMatrix_t<Matrix>;
    using SolverType = mtl::mat::umfpack::solver<FullMatrix>;

  public:
    /// Constructor. Reads UMFPACK parameters from initfile
    UmfpackRunnerBase(std::string prefix)
      : solver(NULL)
      , store_symbolic(0)
      , symmetric_strategy(0)
      , alloc_init(0.7)
    {
      Parameters::get(prefix + "->store symbolic", store_symbolic); // ?
      Parameters::get(prefix + "->symmetric strategy", symmetric_strategy);
      Parameters::get(prefix + "->alloc init", alloc_init);
    }

    /// Destructor
    ~UmfpackRunnerBase() = default;

    /// Implementation of \ref RunnerBase::solve()
    virtual int solve(Matrix const& A, Vector& x, Vector const& b,
                      SolverInfo& solverInfo) override
    {
      AMDIS_FUNCNAME("Umfpack_Runner::solve()");
      test_exit(bool(solver), "The umfpack solver was not initialized\n");

      int code = 0;
      try {
	code = (*solver)(x, b);
      } catch (mtl::mat::umfpack::error& e) {
	error_exit("UMFPACK_ERROR(solve, ", e.code, ") = ", e.what());
      }

      auto r = Vector(b);
      if (two_norm(x) != 0)
        r -= A * x;

      solverInfo.setAbsResidual(two_norm(r));
      solverInfo.setError(code);

      return code;
    }

    /// Implementation of \ref RunnerInterface::adjoint_solve()
    virtual void exit() override {}

  protected:
    shared_ptr<SolverType> solver;

    int store_symbolic;
    int symmetric_strategy;
    double alloc_init;
  };

  // specialization for block-matrices
  template <class SubMatrix, size_t _N, size_t _M, class Vector>
  class UmfpackRunner<BlockMTLMatrix<SubMatrix, _N, _M>, Vector>
    : public UmfpackRunnerBase<BlockMTLMatrix<SubMatrix, _N, _M>, Vector>
  {
    using Matrix = BlockMTLMatrix<SubMatrix, _N, _M>;
    using Super = UmfpackRunnerBase<Matrix, Vector>;

    using SolverType = typename Super::SolverType;

    using Super::symmetric_strategy;
    using Super::alloc_init;
    using Super::solver;

  public:
    UmfpackRunner(std::string prefix)
      : Super(prefix)
    {}

    /// Implementation of \ref RunnerInterface::init()
    virtual void init(Matrix const& A) override
    {
      copy(A, fullMatrix);

      try {
        Super::solver.reset(new SolverType(fullMatrix, symmetric_strategy, alloc_init));
      } catch (mtl::mat::umfpack::error& e) {
        error_exit("UMFPACK_ERROR(factorize, ", e.code, ") = ", e.what());
      }
    }

  private:
    typename Super::FullMatrix fullMatrix;
  };

  // specialization for full-matrices
  template <class T, class Param, class Vector>
  class UmfpackRunner<mtl::compressed2D<T, Param>, Vector>
    : public UmfpackRunnerBase<mtl::compressed2D<T, Param>, Vector>
  {
    using FullMatrix = mtl::compressed2D<T, Param>;
    using Super = UmfpackRunnerBase<FullMatrix, Vector>;

    using SolverType = typename Super::SolverType;

    using Super::symmetric_strategy;
    using Super::alloc_init;
    using Super::solver;

  public:
    UmfpackRunner(std::string prefix)
      : Super(prefix)
    {}

    /// Implementation of \ref RunnerInterface::init()
    virtual void init(FullMatrix const& fullMatrix) override
    {
      try {
        solver.reset(new SolverType(fullMatrix, symmetric_strategy, alloc_init));
      } catch (mtl::mat::umfpack::error& e) {
        error_exit("UMFPACK_ERROR(factorize, ", e.code, ") = ", e.what());
      }
    }
  };
}

#endif // HAVE_UMFPACK
